from enum import Enum

from optimus.engines.base.io.drivers.abstract_driver import AbstractDriver


class DriverContext:
    """Driver context holding a reference to the underlying driver implementation"""

    def __init__(self, driver: AbstractDriver) -> None:
        self._driver = driver

    @property
    def driver(self) -> AbstractDriver:
        return self._driver

    @driver.setter
    def driver(self, driver: AbstractDriver) -> None:
        self._driver = driver

    def properties(self) -> Enum:
        return self._driver.properties()

    def uri(self, *args, **kwargs) -> str:
        return self._driver.uri(*args, **kwargs)

    def url(self, *args, **kwargs) -> str:
        return self._driver.url(*args, **kwargs)

    def table_names_query(self, *args, **kwargs) -> str:
        return self._driver.table_names_query(*args, **kwargs)

    def table_name_query(self, *args, **kwargs) -> str:
        return self._driver.table_name_query(*args, **kwargs)

    def count_query(self, *args, **kwargs) -> str:
        return self._driver.count_query(*args, **kwargs)

    def primary_key_query(self, *args, **kwargs) -> str:
        return self._driver.primary_key_query(*args, **kwargs)

    def min_max_query(self, *args, **kwargs) -> str:
        return self._driver.min_max_query(*args, **kwargs)



import datetime
import numpy as np
from optimus.tests.base import TestBase
from optimus.helpers.json import json_encoding
from optimus.helpers.functions import deep_sort, df_dicts_equal, results_equal


def Timestamp(t):
    return datetime.datetime.strptime(t, "%Y-%m-%d %H:%M:%S")


NaT = np.datetime64('NaT')
nan = float("nan")
inf = float("inf")


class TestExamplePandas(TestBase):
    config = {'engine': 'pandas'}
    dict = {('name', 'object'): ['Optimus', 'Bumblebee', 'Eject'], ('age (M)', 'int64'): [5, 5, 5]}
    maxDiff = None

    def test_cols_upper_multiple(self):
        df = self.df.copy()
        result = df.cols.upper(cols=['name', 'age (M)'])
        result = result.to_dict()
        expected = {'name': ['OPTIMUS', 'BUMBLEBEE', 'EJECT'], 'age (M)': ['5', '5', '5']}
        self.assertEqual(json_encoding(result), json_encoding(expected))

    def test_cols_upper_single(self):
        df = self.df.copy()
        result = df.cols.upper(cols=['name'])
        expected = self.create_dataframe(data={('name', 'object'): ['OPTIMUS', 'BUMBLEBEE', 'EJECT'], ('age (M)', 'int64'): [5, 5, 5]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))


class TestExampleDask(TestExamplePandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestExamplePartitionDask(TestExamplePandas):
    config = {'engine': 'dask', 'n_partitions': 2}


try:
    import cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestExampleCUDF(TestExamplePandas):
        config = {'engine': 'cudf'}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestExampleDC(TestExamplePandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 1}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestExamplePartitionDC(TestExamplePandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 2}


try:
    import pyspark # pyright: reportMissingImports=false
except:
    pass
else:
    class TestExampleSpark(TestExamplePandas):
        config = {'engine': 'spark'}


try:
    import vaex # pyright: reportMissingImports=false
except:
    pass
else:
    class TestExampleVaex(TestExamplePandas):
        config = {'engine': 'vaex'}



from abc import ABC

import numpy as np
import pandas as pd
from fastnumbers import isintlike, isreal, fast_forceint, fast_float

from optimus.engines.base.functions import BaseFunctions
from optimus.helpers.logger import logger
from optimus.infer import is_int_like, is_list_or_tuple


class PandasBaseFunctions(BaseFunctions, ABC):

    @staticmethod
    def is_string(series):
        return series.apply(lambda x: True if type(x) == str else False)

    def is_integer(self, series):
        if str(series.dtype) in self.constants.DATETIME_INTERNAL_TYPES:
            return False
        if str(series.dtype) in self.constants.INT_INTERNAL_TYPES:
            return True
        self.intlike = isintlike
        return np.vectorize(self.intlike)(series).flatten()

    def is_float(self, series):
        if str(series.dtype) in self.constants.DATETIME_INTERNAL_TYPES:
            return False
        # use isreal to allow strings like "0"
        return np.vectorize(isreal)(series).flatten()

    def is_numeric(self, series):
        if str(series.dtype) in self.constants.DATETIME_INTERNAL_TYPES:
            return False
        return np.vectorize(isreal)(series).flatten()


    @classmethod
    def _to_integer(cls, series, default=0):

        # TODO replace_inf

        if is_int_like(default):
            int_type = True
            default = int(default)
            otypes = [int]
        else:
            int_type = False
            otypes = [object]

        try:
            if default is not None:
                series = series.fillna(default)
            series = pd.Series(np.vectorize(fast_forceint,
                                            otypes=otypes)(series, default=default,
                                                           on_fail=lambda x: default).flatten())

        except Exception:
            series = series.replace([np.inf, -np.inf], default)
            if int_type:
                series = pd.Series(np.floor(pd.to_numeric(series, errors='coerce', downcast='integer'))).fillna(default)
                try:
                    series = series.astype('int64')
                except:
                    pass
            else:
                series = pd.Series(np.floor(pd.to_numeric(series, errors='coerce')))
                series = series if default is None else series.fillna(default)

        return series

    # @classmethod
    # def _to_float(cls, series):
    #     try:
    #         return pd.Series(np.vectorize(fast_float)(series, default=np.nan).flatten())
    #     except:
    #         return pd.Series(pd.to_numeric(series, errors='coerce')).astype('float')

    @classmethod
    def _to_datetime(cls, value, format=None):
        try:
            if format is not None:
                return pd.to_datetime(value, format=format, errors="coerce")
        except Exception as e:
            logger.warn(e)
        return pd.to_datetime(value, errors="coerce")

    @classmethod
    def format_date(cls, series, current_format=None, output_format=None):
        return pd.to_datetime(series, format=current_format,
                              errors="coerce").dt.strftime(output_format).reset_index(drop=True)

    @classmethod
    def time_between(cls, series, value=None, date_format=None):

        value_date_format = date_format

        if is_list_or_tuple(date_format) and len(date_format) == 2:
            date_format, value_date_format = date_format

        if is_list_or_tuple(value) and len(value) == 2:
            value, value_date_format = value

        date = pd.to_datetime(series, format=date_format, errors="coerce", utc=True)

        if value is None:
            value = pd.Timestamp.now('utc')
        else:
            value = pd.to_datetime(value, format=value_date_format, errors="coerce", utc=True)

        return (value - date)



from optimus.optimus import Engine, EnginePretty
from optimus.engines.base.engine import BaseEngine

from optimus.engines.cudf.cudf import CUDF
from optimus.engines.cudf.io.load import Load
from optimus._version import __version__

CUDF.instance = None


class CUDFEngine(BaseEngine):
    __version__ = __version__

    def __init__(self, verbose=False, *args, **kwargs):
        import cudf
        self.verbose(verbose)

        CUDF.instance = cudf

        self.client = CUDF.instance

    @property
    def constants(self):
        from optimus.engines.cudf.constants import Constants
        return Constants()

    @property
    def F(self):
        from optimus.engines.cudf.functions import CUDFFunctions
        return CUDFFunctions()

    @property
    def create(self):
        from optimus.engines.cudf.create import Create
        return Create(self)

    @property
    def load(self):
        return Load(self)

    @property
    def engine(self):
        return Engine.CUDF.value
    
    @property
    def engine_label(self):
        return EnginePretty.CUDF.value

    def remote_run(self, callback, *args, **kwargs):
        if kwargs.get("client_timeout"):
            del kwargs["client_timeout"]

        callback(*args, **kwargs)

    def remote_submit(self, callback, *args, **kwargs):
        return self.submit(callback, op=self, *args, **kwargs)

    def submit(self, func, *args, priority=0, pure=False, **kwargs):
        import uuid
        def _func():
            return func(*args, **kwargs)
            
        return type('MockFuture', (object,), {"result": _func, "key": str(uuid.uuid4()), "status": "finished"})



import json

from flask import Flask
from flask import request, render_template

from optimus.server.functions import run, code, engine, session, create_session, delete_session, features

app = Flask(__name__)


@app.route('/<session_key>/run', methods=['POST'])
def post__run(session_key):
    body = json.loads(request.data)
    res = run(session_key, body)
    return json.dumps(res)


@app.route('/<session_key>/code', methods=['POST'])
def post__code(session_key):
    body = json.loads(request.data)
    res = code(session_key, body)
    return json.dumps(res)


@app.route('/<session_key>/init-engine', methods=['POST'])
def post__init_engine(session_key):
    body = json.loads(request.data)
    res = engine(session_key, body or {"engine": "pandas"})
    return json.dumps(res)


@app.route('/<session_key>', methods=['GET'])
def get__session(session_key):
    res = session(session_key)
    return json.dumps(res)


@app.route('/<session_key>', methods=['POST'])
def post__create_session(session_key):
    res = create_session(session_key)
    return json.dumps(res)


@app.route('/<session_key>', methods=['DELETE'])
def delete__session(session_key):
    res = delete_session(session_key)
    return json.dumps(res)


@app.route('/features', methods=['GET'])
def get__features():
    return json.dumps(features())


@app.route('/', methods=['GET'])
def get__playground():
    return render_template('playground.html')




from optimus.helpers.constants import RELATIVE_ERROR
from optimus.outliers.mad import MAD
from optimus.outliers.modified_z_score import ModifiedZScore
from optimus.outliers.tukey import Tukey
from optimus.outliers.z_score import ZScore


class Outliers:
    def __init__(self, df):
        self.df = df

    def tukey(self, cols):
        return Tukey(self.df, cols)

    def z_score(self, cols, threshold):
        return ZScore(self.df, cols, threshold)

    def mad(self, cols, threshold, relative_error=RELATIVE_ERROR):
        return MAD(self.df, cols, threshold, relative_error)

    def modified_z_score(self, cols, threshold, relative_error=RELATIVE_ERROR):
        return ModifiedZScore(self.df, cols, threshold, relative_error)



from enum import Enum

import nltk

from optimus.helpers.logger import logger
from optimus.helpers.raiseit import RaiseIt


class Engine(Enum):
    PANDAS = "pandas"
    CUDF = "cudf"
    DASK = "dask"
    DASK_CUDF = "dask_cudf"
    SPARK = "spark"
    VAEX = "vaex"
    IBIS = "ibis"
    POLARS = "polars"
    PYSCRIPT = "pyscript"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class EnginePretty(Enum):
    PANDAS = "Pandas"
    CUDF = "cuDF"
    DASK = "Dask"
    DASK_CUDF = "Dask-cuDF"
    SPARK = "Spark"
    VAEX = "Vaex"
    IBIS = "Ibis"
    POLARS = "Polars"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


def engine_function(engine, func, *args, **kwargs):
    if engine in Engine.list():
        return func[engine](*args, **kwargs)
    else:
        RaiseIt.value_error(engine, Engine.list())


def start_pandas(*args, **kwargs):
    from optimus.engines.pandas.engine import PandasEngine
    return PandasEngine(*args, **kwargs)


def start_vaex(*args, **kwargs):
    from optimus.engines.vaex.engine import VaexEngine
    return VaexEngine(*args, **kwargs)


def start_spark(*args, **kwargs):
    import databricks.koalas as ks
    """
        Koalas disallows the operations on different DataFrames (or Series) by default to prevent expensive operations.
        It internally performs a join operation which can be expensive in general.
    """
    ks.set_option('compute.ops_on_diff_frames', True)
    from optimus.engines.spark.engine import SparkEngine
    return SparkEngine(*args, **kwargs)


def start_dask(*args, **kwargs):
    from optimus.engines.dask.engine import DaskEngine
    return DaskEngine(*args, **kwargs)


def start_ibis(*args, **kwargs):
    from optimus.engines.ibis.engine import IbisEngine
    return IbisEngine(*args, **kwargs)


def start_cudf(*args, **kwargs):
    from optimus.engines.cudf.engine import CUDFEngine
    return CUDFEngine(*args, **kwargs)


def start_dask_cudf(*args, **kwargs):
    from optimus.engines.dask_cudf.engine import DaskCUDFEngine
    return DaskCUDFEngine(*args, **kwargs)


def start_polars(*args, **kwargs):
    from optimus.engines.polars.engine import PolarsEngine
    return PolarsEngine(*args, **kwargs)


def start_pyscript(*args, **kwargs):
    from optimus.engines.pyscript.engine import PyScriptEngine
    return PyScriptEngine(*args, **kwargs)


def optimus(engine=Engine.DASK.value, *args, **kwargs):
    """
    This is the entry point to initialize the selected engine.
    :param engine: A string identifying an engine :classL`Engine`.
    :param args:
    :param kwargs:
    :return:
    """
    logger.print("ENGINE", engine)

    # lemmatizer
    nltk.download('wordnet', quiet=True)

    # Stopwords
    nltk.download('stopwords', quiet=True)

    # POS
    nltk.download('averaged_perceptron_tagger', quiet=True)

    funcs = {Engine.PANDAS.value: start_pandas,
             Engine.VAEX.value: start_vaex,
             Engine.SPARK.value: start_spark,
             Engine.DASK.value: start_dask,
             Engine.IBIS.value: start_ibis,
             Engine.CUDF.value: start_cudf,
             Engine.DASK_CUDF.value: start_dask_cudf,
             Engine.POLARS.value: start_polars,
             Engine.PYSCRIPT.value: start_polars

             }

    op = engine_function(engine, funcs)

    # Set cupy yo user RMM
    def switch_to_rmm_allocator():
        import rmm
        import cupy
        cupy.cuda.set_allocator(rmm.rmm_cupy_allocator)
        return True

    if engine == Engine.CUDF.value:
        switch_to_rmm_allocator()

    if engine == Engine.DASK_CUDF.value:
        if op.client:
            op.client.run(switch_to_rmm_allocator)

    return op



import os

from optimus.engines.base.io.save import BaseSave, DEFAULT_MODE
from optimus.helpers.functions import prepare_path_local, path_is_local
from optimus.helpers.logger import logger

from optimus.helpers.types import *
from optimus.engines.base.io.save import BaseSave


class Save(BaseSave):
    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def json(self, path, storage_options=None, conn=None, *args, **kwargs):

        df = self.root.data

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            os.makedirs(path, exist_ok=True)
            df.to_json(filename=path, storage_options=storage_options, *args, **kwargs)
        except (OSError, IOError) as error:
            logger.print(error)
            raise

    def csv(self, path, mode=DEFAULT_MODE, index=False, single_file=True, storage_options=None, conn=None, **kwargs):

        df = self.root.data

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            if path_is_local(path):
                prepare_path_local(path)

            df.to_csv(filename=path, mode=mode, index=index, single_file=single_file, storage_options=storage_options,
                      **kwargs)
        except IOError as error:
            logger.print(error)
            raise

    def parquet(self, path, mode=DEFAULT_MODE, num_partitions=1, engine="pyarrow", storage_options=None, conn=None,
                **kwargs):

        # This character are invalid as column names by parquet
        invalid_character = [" ", ",", ";", "{", "}", "(", ")", "\n", "\t", "="]

        def func(col_name):
            for i in invalid_character:
                col_name = col_name.replace(i, "_")
            return col_name

        df = self.root.cols.rename(func)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options
        dfd = df.data
        try:
            if engine == 'pyarrow':
                dfd.to_parquet(path, engine='pyarrow', storage_options=storage_options, **kwargs)
            elif engine == "fastparquet":
                dfd.to_parquet(path, engine='fastparquet', storage_options=storage_options, **kwargs)

        except IOError as e:
            logger.print(e)
            raise

    @staticmethod
    def avro(path):
        raise NotImplementedError('Not implemented yet')



import dask, distributed
from distributed import Client, get_client

from optimus.engines.base.engine import BaseEngine
from optimus.engines.base.remote import MAX_TIMEOUT, RemoteOptimusInterface, RemoteDummyVariable, RemoteDummyDataFrame
from optimus.engines.dask_cudf.io.load import Load
from optimus.helpers.logger import logger
from optimus.optimus import Engine, EnginePretty

BIG_NUMBER = 100000


class DaskCUDFEngine(BaseEngine):
    def __init__(self, session=None, address=None, n_workers=1, threads_per_worker=8, processes=False,
                 memory_limit=None, verbose=False, coiled_token=None, *args, **kwargs):

        """

        :param session:
        :param n_workers:
        :param threads_per_worker:
        :param memory_limit:
        :param verbose:
        :param comm:
        :param args:
        :param kwargs:
        """

        self.engine = Engine.DASK_CUDF.value

        self.verbose(verbose)

        use_remote = kwargs.get("use_remote", coiled_token is not None)

        if coiled_token:
            import coiled
            dask.config.set({"coiled.token": coiled_token})
            try:
                coiled.Cloud()
            except Exception as error:
                raise error

            idle_timeout = kwargs.get("idle_timeout", None)

            memory_limit = memory_limit or '16 GiB'

            cluster = coiled.Cluster(
                name=kwargs.get("name"),
                worker_options={
                    **({"nthreads": threads_per_worker} if threads_per_worker else {})
                },
                worker_gpu=1,
                worker_class='dask_cuda.CUDAWorker',
                n_workers=n_workers if n_workers else 4,
                worker_memory=memory_limit,
                backend_options={
                    "region": kwargs.get("backend_region", "us-east-1")
                },
                scheduler_options={
                    **({"idle_timeout": idle_timeout} if idle_timeout else {})
                },
                software="optimus/gpu"
            )

            self.cluster_name = cluster.name
            self.client = Client(cluster)

        elif address:
            self.client = Client(address=address)

        elif session and session != "local":
            self.client = session

        else:
            try:
                self.client = get_client()
            except ValueError:
                from dask_cuda import LocalCUDACluster
                from GPUtil import GPUtil
                available_gpus = GPUtil.getAvailable(order='first', limit=BIG_NUMBER)
                n_gpus = len(available_gpus)

                if n_workers > n_gpus:
                    logger.print(
                        f"n_workers should be equal or less than the number of GPUs. n_workers is now {n_gpus}")
                    n_workers = n_gpus
                    # n_gpus = 1

                memoryTotal = GPUtil.getGPUs()[available_gpus[0]].memoryTotal

                cluster = LocalCUDACluster(rmm_pool_size=memoryTotal,
                                           device_memory_limit=memoryTotal * 0.8
                                           # Spill to RAM when 80% memory is full
                                           )
                memory_limit = memory_limit or '4GB'
                self.client = Client(cluster, memory_limit=memory_limit, *args, **kwargs)

        if use_remote:
            self.remote = RemoteOptimusInterface(self.client, Engine.DASK_CUDF.value)

        else:
            self.remote = False

    @property
    def constants(self):
        from optimus.engines.base.dask.constants import Constants
        return Constants()

    @property
    def F(self):
        from optimus.engines.dask_cudf.functions import DaskCUDFFunctions
        return DaskCUDFFunctions()

    @property
    def create(self):
        if self.remote:
            return RemoteDummyVariable(self, "_create")

        else:
            from optimus.engines.dask_cudf.create import Create
            return Create(self)

    @property
    def load(self):
        if self.remote:
            return RemoteDummyVariable(self, "_load")

        else:
            return Load(self)

    def dataframe(self, cdf, n_partitions=1, *args, **kwargs):
        import dask_cudf
        from optimus.engines.dask_cudf.dataframe import DaskCUDFDataFrame
        return DaskCUDFDataFrame(dask_cudf.from_cudf(cdf, npartitions=n_partitions, *args, **kwargs))

    def remote_run(self, func, *args, **kwargs):
        if kwargs.get("client_timeout"):
            client_timeout = kwargs.get("client_timeout")
            del kwargs["client_timeout"]
        else:
            client_timeout = MAX_TIMEOUT

        return self.remote_submit(func, *args, **kwargs).result(client_timeout)

    def remote_submit(self, func, *args, **kwargs):

        if not self.remote:
            fut = self.submit(func, op=self, *args, **kwargs)
        else:
            fut = self.remote.submit(func, *args, **kwargs)

        fut.__result = fut.result
        _op = self

        def _result(self, *args, **kwargs):
            result = self.__result(*args, **kwargs)
            if isinstance(result, dict):
                if result.get("status") == "error" and result.get("error"):
                    raise Exception(result.get("error"))
                elif result.get("dummy"):
                    if result.get("dataframe"):
                        return RemoteDummyDataFrame(_op, result.get("dummy"))
                    else:
                        return RemoteDummyVariable(_op, result.get("dummy"))
            return result

        import types
        fut.result = types.MethodType(_result, fut)

        return fut

    def submit(self, func, *args, priority=0, pure=False, **kwargs):
        from optimus.engines.base.remote import RemoteDummyAttribute
        if isinstance(func, (RemoteDummyAttribute,)):
            return func(client_submit=True, *args, **kwargs)
        return distributed.get_client().submit(func, priority=priority, pure=pure, *args, **kwargs)

    @property
    def engine_label(self):
        return EnginePretty.DASK_CUDF.value


import re

import ntpath
from optimus.helpers.raiseit import RaiseIt

import dask.bag as dask_bag
import pandas as pd
from dask import dataframe as dd

import optimus.helpers.functions_spark
from optimus.optimus import EnginePretty
from optimus.engines.base.io.load import BaseLoad
from optimus.engines.base.meta import Meta
from optimus.engines.dask.dataframe import DaskDataFrame
from optimus.helpers.core import val_to_list
from optimus.helpers.functions import prepare_path, unquote_path
from optimus.helpers.logger import logger


class Load(BaseLoad):

    @staticmethod
    def df(*args, **kwargs):
        return DaskDataFrame(*args, **kwargs)

    @staticmethod
    def _csv(filepath_or_buffer, n_partitions=None, nrows=None, engine="c", na_filter=True,
             na_values=None, index_col=None, on_bad_lines='warn', *args, **kwargs):
        
        na_filter = na_filter if na_values else False

        new_dtype = None
        tries = 0

        while tries < 3:

            tries += 1

            if new_dtype is not None:
                logger.warn(f"Load failed, retrying using dtype={new_dtype}")
                if "dtype" in kwargs:
                    kwargs["dtype"].update(new_dtype)
                else:
                    kwargs.update({"dtype": new_dtype})

            if engine == "python":
                on_bad_lines = 'warn'
                df = dd.read_csv(filepath_or_buffer, keep_default_na=True,
                                 na_values=None, engine=engine, on_bad_lines=on_bad_lines, *args, **kwargs)

            elif engine == "c":
                df = dd.read_csv(filepath_or_buffer, keep_default_na=True,
                                 engine=engine, na_filter=na_filter, na_values=val_to_list(na_values),
                                 low_memory=False, on_bad_lines=on_bad_lines, *args, **kwargs)
            
            else:
                RaiseIt.value_error(engine, ["python", "c"])

            try:
                if index_col:
                    df = df.set_index(index_col)

                if nrows:
                    logger.warn(f"'load.csv' on {EnginePretty.DASK.value} loads the whole dataset and then truncates it")
                    df = df.head(n=nrows, compute=False)

                if n_partitions is not None:
                    df = df.repartition(npartitions=n_partitions)

                df = df.persist()

                # TODO use 'check' parameter
                dtypes = df.compute().dtypes
            except ValueError as e:

                if tries >= 3:
                    raise e

                # TODO retry without parsing error message
                e_str = " ".join(str(e).split("\n"))
                found = re.findall(r"dtype=({.*})", e_str)

                if len(found) > 0:
                    new_dtype = eval(found[0])
                else:
                    raise e
            else:
                break

        return df


    @staticmethod
    def _json(filepath_or_buffer, n_partitions=None, *args, **kwargs):
        df = dd.read_json(filepath_or_buffer, *args, **kwargs)

        if n_partitions is not None:
            df = df.repartition(npartitions=n_partitions)
        
        return df

    @staticmethod
    def _avro(filepath_or_buffer, n_partitions=None, nrows=None, *args, **kwargs):
        df = dask_bag.read_avro(filepath_or_buffer, *args, **kwargs).to_dataframe()
        if nrows:
            logger.warn(f"'load.avro' on {EnginePretty.DASK.value} loads the whole dataset and then truncates it")
            df = df.head(n=nrows, compute=False)
        
        if n_partitions is not None:
            df = df.repartition(npartitions=n_partitions)
        
        return df

    @staticmethod
    def _parquet(filepath_or_buffer, n_partitions=None, nrows=None, engine="pyarrow", *args, **kwargs):
        
        df = dd.read_parquet(filepath_or_buffer, engine=engine, *args, **kwargs)

        if nrows:
            logger.warn(f"'load.parquet' on {EnginePretty.DASK.value} loads the whole dataset and then truncates it")
            df = df.head(n=nrows, compute=False)

        if n_partitions is not None:
            df = df.repartition(npartitions=n_partitions)
        
        return df

    def zip(self, path, sep=',', header=True, infer_schema=True, charset="UTF-8", null_value="None", n_rows=-1,
            storage_options=None, conn=None, n_partitions=None, *args, **kwargs):

        path = unquote_path(path)

        file, file_name = prepare_path(path, "zip")

        from zipfile import ZipFile
        import dask.dataframe as dd
        import os

        wd = '/path/to/zip/files'
        file_list = os.listdir(wd)
        destdir = '/extracted/destination/'

        dfd = dd.from_pandas(pd.DataFrame())

        for f in file_list:
            with ZipFile(wd + f, "r") as zip:
                zip.extractall(destdir, None, None)
                df = dd.read_csv(zip.namelist(), usecols=['Enter', 'Columns', 'Here'], parse_dates=['Date'])
                dfd = optimus.helpers.functions_spark.append(df)

        dfd.compute()

        try:
            df = dd.read_csv(file, sep=sep, header=0 if header else None, encoding=charset, na_values=null_value,
                             compression="gzip", *args, **kwargs)

            if n_rows > -1:
                df = df.rows.limit(n_rows)

            df = DaskDataFrame(df, op=self.op)

            df.meta = Meta.set(df.meta, "file_name", file_name)
        except IOError as error:
            logger.print(error)
            raise
        return df

    def orc(self, path, columns=None, cache=None, storage_options=None, conn=None, n_partitions=None, *args, **kwargs):

        path = unquote_path(path)

        if cache is False:
            prepare_path.cache_clear()

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            # From the panda docs using na_filter
            # Detect missing value markers (empty strings and the value of na_values). In data without any NAs,
            # passing na_filter=False can improve the performance of reading a large file.
            dfd = dd.read_orc(path, columns, storage_options=storage_options, *args, **kwargs)

            df = DaskDataFrame(dfd, op=self.op)
            df.meta = Meta.set(df.meta, value={"file_name": path, "name": ntpath.basename(path)})
        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def _xml(filepath_or_buffer, nrows=None, n_partitions=1, *args, **kwargs):
        pdf = pd.read_xml(filepath_or_buffer, *args, **kwargs)
        if nrows:
            logger.warn(f"'load.xml' on {EnginePretty.PANDAS.value} loads the whole dataset and then truncates it")
            pdf = pdf[:nrows]

        df = dd.from_pandas(pdf, npartitions=n_partitions)
        return df


    @staticmethod
    def _excel(path, nrows, storage_options=None, n_partitions=1, *args, **kwargs):
        pdfs = pd.read_excel(path, nrows=nrows, storage_options=storage_options, *args, **kwargs)
        sheet_names = list(pd.read_excel(path, None, storage_options=storage_options).keys())
        pdf = pd.concat(val_to_list(pdfs), axis=0).reset_index(drop=True)
        df = dd.from_pandas(pdf, npartitions=n_partitions)

        return df, sheet_names



import numpy as np
import pandas as pd


from optimus.engines.base.pandas.functions import PandasBaseFunctions
from optimus.engines.base.dataframe.functions import DataFrameBaseFunctions


class PandasFunctions(PandasBaseFunctions, DataFrameBaseFunctions):

    _engine = pd

    @staticmethod
    def dask_to_compatible(dfd):
        from optimus.helpers.converter import dask_dataframe_to_pandas
        return dask_dataframe_to_pandas(dfd)

    @staticmethod
    def df_concat(df_list):
        return pd.concat(df_list, axis=0, ignore_index=True)

    @staticmethod
    def new_df(*args, **kwargs):
        return pd.DataFrame(*args, **kwargs)

    def count_zeros(self, series, *args):
        return int((self.to_float(series).values == 0).sum())

    def kurtosis(self, series):
        # use scipy to match function from dask.array.stats
        from scipy.stats import kurtosis
        return kurtosis(self.to_float(series.dropna()))

    def skew(self, series):
        # use scipy to match function from dask.array.stats
        from scipy.stats import skew
        return skew(self.to_float(series.dropna()))

    def exp(self, series):
        return np.exp(self.to_float(series))

    def sqrt(self, series):
        return np.sqrt(self.to_float(series))

    def reciprocal(self, series):
        return np.reciprocal(self.to_float(series))

    def radians(self, series):
        return np.radians(self.to_float(series))

    def degrees(self, series):
        return np.degrees(self.to_float(series))

    def ln(self, series):
        return np.log(self.to_float(series))

    def log(self, series, base=10):
        return np.log(self.to_float(series)) / np.log(base)

    def sin(self, series):
        return np.sin(self.to_float(series))

    def cos(self, series):
        return np.cos(self.to_float(series))

    def tan(self, series):
        return np.tan(self.to_float(series))

    def asin(self, series):
        return np.arcsin(self.to_float(series))

    def acos(self, series):
        return np.arccos(self.to_float(series))

    def atan(self, series):
        return np.arctan(self.to_float(series))

    def sinh(self, series):
        return np.arcsinh(self.to_float(series))

    def cosh(self, series):
        return np.cosh(self.to_float(series))

    def tanh(self, series):
        return np.tanh(self.to_float(series))

    def asinh(self, series):
        return np.arcsinh(self.to_float(series))

    def acosh(self, series):
        return np.arccosh(self.to_float(series))

    def atanh(self, series):
        return np.arctanh(self.to_float(series))

    def floor(self, series):
        return np.floor(self.to_float(series))

    def ceil(self, series):
        return np.ceil(self.to_float(series))
        
    def normalize_chars(self, series):
        return self.to_string_accessor(series).normalize("NFKD").str.encode('ascii', errors='ignore').str.decode('utf8')



import timeit

from optimus.helpers.logger import logger


def time_it(method):
    def timed(*args, **kw):
        start_time = timeit.default_timer()
        f = method(*args, **kw)
        _time = round(timeit.default_timer() - start_time, 2)
        logger.print("{name}() executed in {time} sec".format(name=method.__name__, time=_time))
        return f

    return timed


def apply_to_categories(func):

    if hasattr(func, "__func__"):
        func = func.__func__
        static = True
    else:
        static = False

    def wrapper(ref, series, *args, **kwargs):
        # TODO:
        #  Handle Ibis Series. Ibis use 'type' instead of dtype.
        #  This implementation should be handle in the Optimus Ibis Dataframe

        # if series doesn't have a 'known' property it is a dask series,
        # which already optimizes this kind of operations
        if hasattr(series, "dtype"):
            if str(series.dtype) == "category" and hasattr(series, "cat") and not hasattr(series.cat, "known"):

                transformed = series.cat.categories.to_series()
                if static:
                    transformed = func(transformed, *args, **kwargs)
                else:
                    transformed = func(ref, transformed, *args, **kwargs)
                string_index = series.cat.categories.__class__(ref.to_list(transformed))
                return series.cat.set_categories(string_index, rename=True)

        if static:
            return func(series, *args, **kwargs)
        else:
            return func(ref, series, *args, **kwargs)

    return wrapper
